import ast
import imp
import inspect
import types
from .invocation_tools import _name_of


class PipelineCommand:
    def __init__(self, name):
        self._name = name


class SkipTo(PipelineCommand):
    # TODO: Singleton?
    def __init__(self, resuming_function_name):
        self.function_name = _name_of(resuming_function_name)

Done = PipelineCommand('DONE')


def collect_func_ordering(file_path_or_module):
    """
    Collect all top-level functions in a file or module, in order.

    :param file_path_or_module: the path to the python file, or a module.
    :return: the ordered top-level function names.
    """
    if isinstance(file_path_or_module, types.ModuleType):
        file_path = inspect.getsourcefile(file_path_or_module)
    else:
        file_path = file_path_or_module

    with open(file_path, 'r') as fp:
        s = fp.read()
        print(type(file_path_or_module))
        print("]]]]" + s + "[[[[")

        root = ast.parse(s)

    names = []
    for node in ast.iter_child_nodes(root):
        if isinstance(node, ast.FunctionDef):
            names.append(node.name)
    return names


def collect_pipeline(module, skip_private=True, reloading=True):
    """
    Load the functions in a module in their definition order.

    :param module: a python module
    :param skip_private: ignore functions starting with '_'
    :param reloading: reload the module prior to collection
    :return: the functions in their definition order
    """
    if reloading:
        module = imp.reload(module)

    pipeline = []
    env = vars(module)
    for name in collect_func_ordering(module):
        if skip_private and name.startswith('_'):
            continue
        pipeline.append(env[name])
    return pipeline


class Pipeline(object):
    def __init__(self):
        self._pipeline = []
        self._captures = []

    def __iter__(self):
        return iter(self._captures)

    def __call__(self, *args, **kwargs):
        await_f = None
        for f in self._pipeline:
            if await_f is not None:
                if _name_of(f) != await_f:
                    continue
                else:
                    await_f = None

            cmd = f(*args, **kwargs)

            if isinstance(cmd, PipelineCommand):
                if cmd is Done:
                    break
                elif isinstance(cmd, SkipTo):
                    await_f = cmd.function_name

        if await_f is not None:
            raise NameError("Function {} never visited".format(await_f))


class ModulePipeline(Pipeline):
    """
    Extract a pipeline from a Python module.

    This executes each function in the order laid out by the file.
    """
    def __init__(self, module, skip_private_applications=True,
                 include_private_captures=True, reloading=True):
        """

        :param module: the python module to load
        :param skip_private_applications: if True, then functions prefixed with
            '_' are not included in the transformation pipeline
        :param include_private_captures: if True, then functions prefixed with
            '_' ARE captured for error analysis.
        :param reloading: if True, reloads the module when calling `reload()`
        """
        self._module = module
        self._skip_private_applications = skip_private_applications
        self._include_private_captures = include_private_captures
        self._reloading = reloading

        super(ModulePipeline, self).__init__()

        self.reload(force=True)

    def reload(self, force=False):
        """
        Reload the underlying module.

        :param force: if True, reloads the module, even if reloading is false.
        """
        # Note: reloading the module on the initial load actually makes sense
        # given how it's used. In a notebook, you import the module, then
        # pass it to the constructor. It's easy to step over that
        # constructor again, passing the old module reference.
        if force or self._reloading:
            self._pipeline = []
            self._captures = []
            for f in collect_pipeline(self._module, skip_private=False):
                is_private = f.__name__.startswith('_')

                if not self._skip_private_applications or not is_private:
                    self._pipeline.append(f)

                if self._include_private_captures or not is_private:
                    self._captures.append(f)



import re


def rename_ks(d, mapping, must_exist=False):
    """
    Rename the keys in a given dictionary.

    :param d: the dict to update
    :param mapping: a dict of old_key -> new_key pairs
    """
    for k_old, k_new in mapping.items():
        key_exists = k_old in d
        if key_exists:
            d[k_new] = d[k_old]
            del d[k_old]
        elif must_exist:
            assert key_exists, "Key {} not in {}".format(k_old, d)


PYTHONIZE_IDENTIFIER_RE = re.compile('[^0-9a-z]+')


def pythonize_identifier(s):
    s = PYTHONIZE_IDENTIFIER_RE.sub('_', s.lower())
    if s and s[0] == '_':
        s = s[1:]
    if s and s[-1] == '_':
        s = s[:-1]
    return s


def pythonize_ks(d, recursive=True):
    remapping = {}
    for k, v in d.items():
        pythonized = pythonize_identifier(k)
        if pythonized != k:
            remapping[k] = pythonized

        if recursive:
            if isinstance(v, dict):
                pythonize_ks(v)

    rename_ks(d, remapping)


def fill_in_unknowns(src_doc, dst_doc):
    """
    Recursively update dst_doc with items from the src_doc if it doesn't exist.

    :param src_doc: a dict with sub-documents
    :param dst_doc: a dict with sub-documents
    """
    for k, v in src_doc.items():
        if isinstance(v, dict):
            if k in dst_doc:
                assert isinstance(dst_doc[k], dict)
            else:
                dst_doc[k] = {}
            fill_in_unknowns(v, dst_doc[k])
        else:
            if k not in dst_doc:
                dst_doc[k] = v


def update_values(d, ks, converter, must_exist):
    for k in ks:
        key_exists = k in d
        if key_exists:
            d[k] = converter(d[k])
        elif must_exist:
            assert key_exists, "Key {} not in {}".format(k, d)


def dict_values_to_int(d, ks, must_exist=False):
    """
    Convert values of a dict to ints for some keys.

    :param d: a dict
    :param ks: an iterable of keys
    :param must_exist: raises an exception if set to True and the given key
        does not exist in the given dict
    """
    update_values(d, ks, int, must_exist)


def dict_values_to_float(d, ks, must_exist=False):
    """
    Convert values of a dict to floats for some keys.

    :param d: a dict
    :param ks: an iterable of keys
    :param must_exist: raises an exception if set to True and the given key
        does not exist in the given dict
    """
    update_values(d, ks, float, must_exist)


def dict_values_to_upper_if_str(d):
    for k, v in d.items():
        if isinstance(v, str):
            d[k] = v.upper()


COMMON_BOOL_TRUE = set(['yes', 'true', 'y', 't', '1'])
COMMON_BOOL_FALSE = set(['no', 'false', 'n', 'f', '0'])


def str_to_bool(s, true_set=COMMON_BOOL_TRUE, false_set=COMMON_BOOL_FALSE):
    """
    Convert a given string to a boolean value.

    :param s: a boolean value represented as a string
    :param true_set: a set of (all-lowercase) values indicating true
    :param false_set: a set of (all-lowercase) values indicating false
    :return: True if the value is in `true_set`, False if it is in `false_set`
        otherwise `None`
    """
    if isinstance(s, str):
        s = s.lower()
        if s in true_set:
            return True
        elif s in false_set:
            return False

    return None


def dict_string_values_to_bool(d, ks, true_set=None, false_set=None,
                               must_exist=False):
    """
    Convert values of a dict to floats for some keys.

    :param d: a dict
    :param ks: an iterable of keys
    :param must_exist: raises an exception if set to True and the given key
        does not exist in the given dict
    """
    true_set = true_set or COMMON_BOOL_TRUE
    false_set = false_set or COMMON_BOOL_FALSE
    f = lambda s: str_to_bool(s, true_set, false_set)
    update_values(d, ks, f, must_exist)


WHITESPACE_RE = re.compile("\s+")


def sstrip(s):
    """
    :param s: a string
    :return: a new string with prefix and postfix spacing removed, and all
        all spacing converted to single spacing
    """
    return WHITESPACE_RE.sub(' ', s).strip()


def update_items_in_place(items, conversion):
    for i, item in enumerate(items):
        items[i] = conversion(item)


def sstrip_all(items):
    """
    Applying sstrip to items, updating in-place.
    """
    update_items_in_place(items, sstrip)


def flat_mapping_of_dict(d, sep='_', prefix=''):
    """
    Return a copy of d, recursively unpacking all dicts.

    Note: This assumes all keys are strings!

    :param d: a dict
    :param sep: the prefix to a key
    :param prefix: the prefix to keys in this dict
    :return: a new dict
    """
    new_d = {}
    for k, v in d.items():
        k = prefix + sep + str(k) if prefix else str(k)

        if isinstance(v, dict):
            new_d.update(flat_mapping_of_dict(v, sep=sep, prefix=k))
        else:
            new_d[k] = v
    return new_d


def remove_private_keys(d, recursive=True, prefix='_'):
    ks = []

    for k, v in d.items():
        if k.startswith(prefix):
            ks.append(k)

        if isinstance(v, dict):
            remove_private_keys(v, recursive, prefix)

    for k in ks:
        del d[k]



import os
import json
import numpy as np
import pandas as pd
from fnmatch import fnmatch
from functools import reduce


def first(items):
    """
    Get the first item from an iterable.

    Warning: It consumes from a generator.

    :param items: an iterable
    :return: the first in the iterable
    """
    for item in items:
        return item


def jsonlines_reader(file_path, skip_decode_errors=False):
    """
    Yield each document in a JSON-lines document.
    :param file_path: filepath to a json-lines file set
    """
    with open(file_path) as fp:
        for line in fp:
            try:
                yield json.loads(line)
            except json.JSONDecodeError:
                if not skip_decode_errors:
                    raise


def files_processor(generator_func, dir_path, shell_ptn="*", recursive=False):
    for file_name in os.listdir(dir_path):
        file_path = os.path.join(dir_path, file_name)
        if recursive and os.path.isdir(file_path):
            for item in files_processor(generator_func, file_path,
                                        shell_ptn, True):
                yield item
        else:
            if fnmatch(file_name, shell_ptn):
                for item in generator_func(file_path):
                    yield item


def pd_print_entirely(frame_or_series):
    """
    Print a pandas dataframe or series in its entirety.

    :param frame_or_series: a pandas Series or DataFrame
    """
    columns = pd.get_option('display.max_columns')
    pd.set_option('display.max_columns', None)

    rows = pd.get_option('display.max_rows')
    pd.set_option('display.max_rows', None)

    try:
        print(frame_or_series)
    finally:
        pd.set_option('display.max_columns', columns)
        pd.set_option('display.max_rows', rows)


def np_and(condition, *conditions):
    # This is just much less noise then (a) && (b) && (c) to me.
    return reduce(np.logical_and, conditions, condition)


def np_or(condition, *conditions):
    # This is just much less noise then (a) || (b) || (c) to me.
    return reduce(np.logical_or, conditions, condition)




import jmespath
import random

# See also: fill_in_unknowns in transformations as a collector.


def sampling(items, p):
    for item in items:
        if random.random() < p:
            yield item


class Collector:
    def update(self, item):
        raise NotImplementedError("Collector#update(item) not implemented")

    def update_over_all(self, items):
        for item in items:
            self.update(item)

    @property
    def collected(self):
        raise NotImplementedError("Collector#collected not implemented")


class SetCollector(Collector):
    """
    Collect the set of values for jmespaths over applied items.
    """
    def __init__(self, paths):
        self._paths = {}
        self._sets = {}
        self.add_paths(paths)

    def add_paths(self, paths):
        """
        :param paths: an interable of jmespath paths
        """
        for path in paths:
            self.add_path(path)

    def add_path(self, path):
        """
        :param path: a jmespath
        """
        self._paths[path] = jmespath.compile(path)

    def update(self, item):
        """
        Apply the paths to an item, collecting the values.

        :param item: an item to process
        """
        for path, jmes_obj in self._paths.items():
            res = jmes_obj.search(item)
            if res is not None:
                result_set = self._sets.get(path)
                if not result_set:
                    result_set = set()
                    self._sets[path] = result_set
                result_set.add(res)

    @property
    def collected(self):
        return self._sets


class GroupCollector(Collector):
    """
    Collect one item per group.
    """

    def __init__(self, group_f):
        """
        :param group_f: function which returns some key representing the group
        """
        self._group_f = group_f
        self._groups = {}

    def update(self, item):
        k = self._group_f(item)
        if k not in self._groups:
            self._groups[k] = item

    @property
    def collected(self):
        return self._groups



import os
import types


def _clean_filename(filename):
    """
    Make a traceback code-frame's file name more human readable.

    :param filename: the file name given in a traceback code frame.
    :return: the human-readable file name
    """
    # The file name for an IPython input cell looks like:
    #    <ipython-input-111-8515003c2eb6>
    # It'd be less noisy to read in input cell format:
    #    In [111]
    if filename.startswith("<ipython-input-"):
        return "In [{}]".format(filename.split("-")[2])
    else:
        return os.path.basename(filename)


def _extract_args(frame):
    """
    Extract the arguments given to the function called in the given frame.

    :param frame: the frame from a traceback object
    :return: None if there were no arguments, otherwise a tuple of the
        function's arguments.
    """
    # Number of arguments given to the function.
    # But, this doesn't include * or ** arguments.
    code = frame.f_code
    n_args = code.co_argcount
    if n_args == 0:
        return None
    else:
        # The `co_varnames` object seems like a stack, listing variables in
        # creation order. The first `n_args` bindings are the function's
        # arguments.
        #
        # Warning: Don't make this a tuple, even though it feels natural.
        # JMESPath doesn't like array-indexing on tuples.
        return [frame.f_locals.get(k) for k in code.co_varnames[:n_args]]


def no_filter(invocation_dict):
    return True


def _extract_invocations(tb, accept_pred=no_filter):
    """
    Extract a list of function invocation information from a traceback.

    :param tb: the traceback object
    :param accept_pred: given the extracted invocation details, this
        predicate returns True to include this in the results list. The
        default predicate does no filtering.
    :return: a list of dictionaries with `name`, `file_name`, `lineno`,
        `call_args` keys associated with the invocation's traceback frame
    """
    invocations = []

    while tb:
        frame = tb.tb_frame
        code = frame.f_code

        invocation = {'name': code.co_name,
                      'filename': _clean_filename(code.co_filename),
                      'lineno': tb.tb_lineno,
                      'call_args': _extract_args(frame)}
        if accept_pred(invocation):
            invocations.append(invocation)
        tb = tb.tb_next

    return invocations


def callables_from(env, include_private=False):
    """
    Collect all the callable objects in the given environment

    :param env: the target environment (i.e. binding -> value dict). Usually,
        this should be `globals()`. If it is an object, it collects all methods
        in that object. If it is a module, it collects all top-level callables.
    :return: a list of the callable objects' names
    """
    if isinstance(env, types.ModuleType):
        env = vars(env)
    elif not isinstance(env, dict):
        # TODO: BUG: Will this break descriptors and property methods?
        env = {k: getattr(env, k) for k in dir(env)}

    if include_private:
        return [name for name, obj in env.items() if hasattr(obj, '__call__')]
    else:
        return [name for name, obj in env.items()
                if hasattr(obj, '__call__') and not name.startswith('_')]




def _name_of(f):
    return f.__name__ if hasattr(f, '__name__') else f



from collections import defaultdict

import jmespath

from .invocation_tools import _extract_invocations, _name_of


class VaqueroException(Exception):
    pass


class Vaquero:
    def __init__(self, max_failures=None, capture_error_invocations=True,
                 annotate_failure_with_input=False):
        """
        :param max_failures: The maximum number of failures allowed before
            loudly signaling a problem. The default is None, which means the
            Vaquero object never raises an exception.

        :param capture_error_invocations: if True (default) captures errant
            invocations for post-processing. Otherwise, doesn't capture
            (useful for more production-oriented code)

        :param capture_error_invocations: if True (not default) then the
            captured invocation is annotated with a `top_input` field
            set to the input (as set by set_input)
        """
        self._max_failures = max_failures
        self._capture_error_invocations = capture_error_invocations
        self._annotate_failure_with_input = annotate_failure_with_input
        self._target_funcs = set()
        self.reset()

    def reset(self, turn_off_error_capturing=False):
        """
        Resets all internal counters and captured exceptions.

        :param turn_off_error_capturing:
        :return:
        """
        self._successes = 0
        self._failures = 0
        self._captures = defaultdict(list)
        self._last_input = None

        if turn_off_error_capturing is True:
            self._capture_error_invocations = False

    def stats(self):
        """
        :return: a dict with `successes`, `failures`, `ignored`, and
            `failures_by`
        """
        return {'successes': self._successes,
                'failures': self._failures,
                'ignored': 0,
                'failures_by': {k: len(v) for k, v in self._captures.items()}}

    @property
    def n_failures(self):
        return self._failures

    def merge(self, other):
        """
        Merge this Vaquero object with another.

        :param other: another vaquero object
        """
        self._successes += other._successes
        self._failures += other._failures

        for k, v in other._captures.items():
            if k in self._captures:
                self._captures[k].extend(v)
            else:
                self._captures[k] = v

    def stats_ignoring(self, *exc_types_to_ignore):
        """
        Works like `stats` but ignores some captured failures.

        :param exc_types_to_ignore: an arbitrary set of `exc_type`s to ignore.
        """
        exc_types_to_ignore = set([t.__name__ if hasattr(t, '__name__') else t
                                   for t in exc_types_to_ignore])

        n_ignored, n_failures, failures_by = 0, self._failures, {}
        for k, captures in self._captures.items():
            failures_by_k = 0
            for capture in captures:
                if capture['exc_type'] in exc_types_to_ignore:
                    n_failures -= 1
                    n_ignored += 1
                else:
                    failures_by_k += 1
            if failures_by_k > 0:
                failures_by[k] = failures_by_k

        return {'successes': self._successes,
                'ignored': n_ignored,
                'failures': n_failures,
                'failures_by': failures_by}

    def register_target(self, f):
        """
        Registers a function with the exception capturing machinery.

        :param f: a function (or string naming a function)
        """
        self._target_funcs.add(_name_of(f))

    def register_targets(self, target_funcs):
        """
        Register functions with the exception capturing machinery.

        :param target_funcs: a collection of functions or string names
        """
        for f in target_funcs:
            self.register_target(f)

    @property
    def target_funcs(self):
        return set(self._target_funcs)

    def set_input(self, input):
        """
        Set the input **prior** to using it.

        Useful for post failure analysis via the `last_input` property.
        """
        self._last_input = input

    def on_input(self, input):
        """
        Set the input **prior** to using it, then return the context.
        """
        self.set_input(input)
        return self

    @property
    def was_successful(self):
        return self._successes > 0 and self._failures == 0

    @property
    def last_input(self):
        return self._last_input

    def __enter__(self):
        return self.set_input

    def __exit__(self, exc_type, exc_value, tb):
        # TODO only catch certain types
        if exc_type is None:
            self._successes += 1
        elif exc_type == KeyboardInterrupt:
            return True  # Swallow exception. Programmer error!
        else:
            self._failures += 1

            # Gather invocations only for registered functions.
            invocations = [invoc for invoc in _extract_invocations(tb)
                           if invoc['name'] in self._target_funcs]
            if invocations:
                root_cause = invocations[-1]

                if self._capture_error_invocations:
                    if self._annotate_failure_with_input:
                        root_cause['top_input'] = self._last_input

                    # These are string-encoded for ease of JMES use.
                    root_cause['exc_type'] = exc_type.__name__
                    root_cause['exc_value'] = str(exc_value)

                    self._captures[root_cause['name']].append(root_cause)

                if self._failures > (self._max_failures or self._failures):
                    raise VaqueroException("Failure limit exceeded!")

                return True  # Swallow exception: see PEP-0343
            else:
                raise  # Exception in non-registered code. Don't swallow!

    def examine(self, f, selector=None, as_set=False):
        """
        Examine the exceptional cases for a given function.

        :param f: the function or name to examine
        :param selector: the JMES-based selector to apply to the captures
        :param as_set: if True return the selection result as a set. Only
            applies to examinations with a selector.
        :return: the invocations or a transformation of invocations
        """
        captures = self._captures[_name_of(f)]

        if selector is None:
            assert not as_set, "as_set=True only works with a selector"
            return captures
        else:
            items = jmespath.search(selector, captures)
            return set(items) if as_set else items

    def first_failure(self, f):
        errors = self.examine(f)
        assert errors, "No failure on {}".format(f)
        return errors[0]

    def execute_on_failure(self, f, i, captured_func_name=None):
        """
        Execute's f on some captured failure.

        :param f: a function that takes the captured invocation for its own
            name or the `captured_func_name` override
        :param i: the index of the capture for this function
        :param captured_func_name: override the name of the capture
        :return: the result of f(*captured_args)
        """
        name = captured_func_name or _name_of(f)
        if name not in self._captures:
            raise VaqueroException("Function `{}` has no issues".format(name))
        return f(*self._captures[name][i]['call_args'])

    def execute_over_failures(self, f, captured_func_name=None, *args, **kwargs):
        """
        Applies each set of failing arguments for f to a new vaquero object.

        :param f: the modified version of a registered callable
        :param captured_func_name: if specified, this is the captures to use
            for f, rather than inferring the name from f. It's useful if you
            want to experiment without overriding the original function
        :param args: passed to a new Vaquero object
        :param kwargs: passed to a new Vaquero object
        :return: the result of a Vaquero object capturing f(*args_in_failure)
        """
        name = captured_func_name or _name_of(f)
        if name not in self._captures:
            raise VaqueroException("Function `{}` has no issues".format(name))

        sub_vaquero = Vaquero(*args, **kwargs)
        sub_vaquero.register_target(f)

        for capture in self._captures[name]:
            args = capture['call_args']
            with sub_vaquero:
                f(*args)

        return sub_vaquero




from setuptools import setup, find_packages


setup(
    name='vaquero',
    version='0.0.1',
    description='A tool for iterative and interactive data wrangling.',
    long_description='See: `github repo <https://github.com/jbn/vaquero>`_.',
    url='https://github.com/jbn/vaquero',
    author='John Bjorn Nelson',
    author_email='jbn@pathdependent.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    keywords=['data analysis', 'data science', 'elt'],
    packages=find_packages(exclude=['contrib', 'docs', 'tests*'])
)
